from decimal import Decimal

import pytest

from django_ctb import models as m
from django_ctb.mouser.services import MouserService
from django_ctb.mouser.client import MouserClient, MouserPart, MouserPricebreak


class TestMouserService:
    def test__populate(self, monkeypatch, vendor_part_mouser):
        mouser_part = MouserPart(
            description="Fake part",
            name="BIGBOI1234",
            mouser_part_number="233-FAKE",
            url_path="https://www.mouser.com/whatever/path",
            price_breaks=[
                MouserPricebreak(volume=1, cost="$0.01"),
                MouserPricebreak(volume=10, cost="$0.009"),
                MouserPricebreak(volume=100, cost="$0.008"),
            ],
        )

        def fake_client_get_part(*args, **kwargs):
            return mouser_part

        monkeypatch.setattr(MouserClient, "get_part", fake_client_get_part)
        MouserService()._populate(vendor_part_mouser)
        vendor_part_mouser.refresh_from_db()
        assert vendor_part_mouser.url_path == "/whatever/path"
        assert vendor_part_mouser.volume == 10
        assert 0.00899 < vendor_part_mouser.cost < 0.00901
        assert vendor_part_mouser.part.name == "BIGBOI1234"
        assert vendor_part_mouser.part.value == "BIGBOI1234"
        assert vendor_part_mouser.part.description == "Fake part"

    def test_populate(self, monkeypatch, vendor_part_mouser):
        call_count = 0

        def fake__populate(self, vendor_part):
            assert vendor_part_mouser == vendor_part
            nonlocal call_count
            call_count += 1

        monkeypatch.setattr(MouserService, "_populate", fake__populate)
        MouserService().populate(vendor_part_mouser.pk)
        assert call_count == 1

    def test_populate_missing(self, db, monkeypatch):
        call_count = 0

        def fake__populate(self, vendor_part):
            assert vendor_part_mouser == vendor_part
            nonlocal call_count
            call_count += 1

        monkeypatch.setattr(MouserService, "_populate", fake__populate)
        MouserService().populate(123)
        assert call_count == 0

    def test__create_missing_part(self, monkeypatch):
        def fake_client_get_part(*args, **kwargs):
            raise MouserClient.EmptyResponse

        monkeypatch.setattr(MouserClient, "get_part", fake_client_get_part)
        with pytest.raises(MouserService.MissingPart):
            ret = MouserService()._create("asdf-1234")

    def test__create(self, db, monkeypatch, vendor_mouser):
        assert m.Part.objects.all().count() == 0
        assert vendor_mouser.vendor_parts.all().count() == 0
        mouser_part = MouserPart(
            description="Fake part",
            name="BIGBOI1234",
            mouser_part_number="233-FAKE",
            url_path="https://www.mouser.com/whatever/path",
            price_breaks=[
                MouserPricebreak(volume=1, cost="$0.01"),
                MouserPricebreak(volume=10, cost="$0.009"),
                MouserPricebreak(volume=100, cost="$0.008"),
            ],
        )

        def fake_client_get_part(*args, **kwargs):
            return mouser_part

        monkeypatch.setattr(MouserClient, "get_part", fake_client_get_part)
        ret = MouserService()._create("asdf-1234")
        assert m.Part.objects.all().count() == 1
        p = m.Part.objects.get()
        assert p.name == "BIGBOI1234"
        assert p.value == "BIGBOI1234"
        assert p.description == "Fake part"
        assert vendor_mouser.vendor_parts.all().count() == 1
        vp = vendor_mouser.vendor_parts.get()
        assert vp.part == p
        assert vp.item_number == "asdf-1234"
        assert vp.volume == 10
        assert vp.cost == Decimal("0.0090")
        assert vp.url_path == "/whatever/path"
        vp.delete()
        p.delete()



from django.contrib import admin
from django.shortcuts import render
from functools import update_wrapper
from django.utils.encoding import force_str
from django.http import Http404
from django.contrib.admin.utils import unquote
from django.urls import path, reverse

from django_ctb.tasks import (
    sync_project_version,
    clear_to_build,
    complete_build,
    cancel_build,
    complete_order,
    populate_mouser_vendor_part,
)

from django_ctb import models


@admin.register(models.Footprint)
class FootprintAdmin(admin.ModelAdmin):
    list_display = ("name",)


class ImplicitProjectPartInline(admin.TabularInline):
    # fields = ("item_number", "cost", "volume", "url_path")
    model = models.ImplicitProjectPart


@admin.register(models.Package)
class PackageAdmin(admin.ModelAdmin):
    list_display = ("name", "technology")
    list_filter = ("technology",)
    inlines = (ImplicitProjectPartInline,)


@admin.register(models.Vendor)
class VendorAdmin(admin.ModelAdmin):
    list_display = ("name",)


class VendorOrderLineInline(admin.TabularInline):
    model = models.VendorOrderLine


@admin.register(models.VendorOrder)
class VendorOrderAdmin(admin.ModelAdmin):
    list_display = ("vendor", "order_number", "created", "fulfilled")
    inlines = (VendorOrderLineInline,)
    actions = ("_complete_order",)

    def _complete_order(self, request, queryset):
        for row in queryset:
            complete_order.send(row.pk)
        self.message_user(request, f"{len(queryset)} processes started")

    _complete_order.short_description = "Mark order fulfilled"


class VendorPartInline(admin.TabularInline):
    # fields = ("item_number", "cost", "volume", "url_path")
    model = models.VendorPart
    extra = 1


class InventoryLineInline(admin.TabularInline):
    # fields = ("item_number", "cost", "volume", "url_path")
    model = models.InventoryLine
    extra = 1


@admin.register(models.Part)
class PartAdmin(admin.ModelAdmin):
    list_display = ("description", "symbol", "value", "package")
    list_filter = ("symbol", "name", "package", "value")
    inlines = [VendorPartInline, InventoryLineInline]


@admin.register(models.VendorPart)
class VendorPartAdmin(admin.ModelAdmin):
    list_display = ("item_number", "vendor", "part")
    list_filter = ("vendor",)
    actions = ("_populate",)

    def _populate(self, request, queryset):
        for row in queryset:
            populate_mouser_vendor_part.send(row.pk)
        self.message_user(request, f"{len(queryset)} processes started")

    _populate.short_description = "Populate fields (Mouser)"


@admin.register(models.ImplicitProjectPart)
class ImplicityProjectPartAdmin(admin.ModelAdmin):
    list_display = ("part", "for_package", "quantity")


@admin.register(models.Inventory)
class InventoryAdmin(admin.ModelAdmin):
    list_display = ("name",)


class InventoryActionInline(admin.TabularInline):
    fields = ("delta", "order_line", "build", "created")
    model = models.InventoryAction

    extra = 0


@admin.register(models.InventoryLine)
class InventoryLineAdmin(admin.ModelAdmin):
    list_display = ("part", "quantity", "inventory")
    list_filter = ("inventory", "part__symbol", "part__package__name", "part__value")
    inlines = [InventoryActionInline]


class ProjectVersionInline(admin.TabularInline):
    model = models.ProjectVersion

    extra = 1


@admin.register(models.Project)
class ProjectAdmin(admin.ModelAdmin):
    list_display = ("name", "git_url")
    inlines = (ProjectVersionInline,)


class ProjectBuildInline(admin.TabularInline):
    fields = ("quantity", "created")
    model = models.ProjectBuild


class MissingPartInline(admin.TabularInline):
    fields = ("part", "missing_part_description")
    model = models.ProjectPart
    extra = 0

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.filter(part__isnull=True)


@admin.register(models.ProjectVersion)
class ProjectVersionAdmin(admin.ModelAdmin):
    list_display = ("project", "revision", "commit_ref", "synced", "missing_part_count")
    list_filter = ("project",)
    inlines = [MissingPartInline, ProjectBuildInline]
    actions = ("sync_bom",)

    def sync_bom(self, request, queryset):
        for row in queryset:
            sync_project_version.send(row.pk)
        self.message_user(request, f"{len(queryset)} processes started")

    sync_bom.short_description = "Sync selected version BOMs"

    def missing_part_count(self, obj):
        return obj.project_parts.filter(part__isnull=True).count()

    def _getobj(self, request, object_id):
        opts = self.model._meta

        try:
            obj = self.get_queryset(request).get(pk=unquote(object_id))
        except self.model.DoesNotExist:
            # Don't raise Http404 just yet, because we haven't checked
            # permissions yet. We don't want an unauthenticated user to
            # be able to determine whether a given object exists.
            obj = None

        if obj is None:
            raise Http404(
                f"{force_str(opts.verbose_name)} object with primary key "
                f"'{force_str(object_id)}' does not exist."
            )

        return obj

    def _wrap(self, view):
        def wrapper(*args, **kwargs):
            return self.admin_site.admin_view(view)(*args, **kwargs)

        return update_wrapper(wrapper, view)

    def _view_name(self, name):
        return f"{self.model._meta.app_label}_{self.model._meta.model_name}_{name}"

    def get_urls(self):
        urls = super().get_urls()

        my_urls = [
            path(
                "<object_id>/bom/",
                self._wrap(self.bom_view),
                name=self._view_name("bom"),
            ),
        ]

        return my_urls + urls

    def bom_view(self, request, object_id):
        return render(
            request,
            "admin/inventory/project_version_bom.html",
            {"project_version": self._getobj(request, object_id)},
        )


class ProjectPartFootprintRefInline(admin.TabularInline):
    fields = ("footprint_ref",)
    model = models.ProjectPartFootprintRef


@admin.register(models.ProjectPart)
class ProjectPartAdmin(admin.ModelAdmin):
    list_display = ("project_version", "line_number", "part", "quantity")
    list_filter = ("project_version", "part")
    inlines = [ProjectPartFootprintRefInline]


class ProjectBuildPartShortageInline(admin.TabularInline):
    fields = ("part", "quantity")
    model = models.ProjectBuildPartShortage
    extra = 0


@admin.register(models.ProjectBuild)
class ProjectBuildAdmin(admin.ModelAdmin):
    list_display = (
        "project_version",
        "created",
        "quantity",
        "shortfalls",
        "cleared",
        "completed",
    )
    list_filter = ("project_version",)
    date_hierarchy = "created"
    inlines = [ProjectBuildPartShortageInline]
    actions = ("_clear_to_build", "_complete_build", "_cancel_build")

    def _clear_to_build(self, request, queryset):
        for row in queryset:
            clear_to_build.send(row.pk)
        self.message_user(request, f"{len(queryset)} processes started")

    _clear_to_build.short_description = "Clear to build"

    def _complete_build(self, request, queryset):
        for row in queryset:
            complete_build.send(row.pk)
        self.message_user(request, f"{len(queryset)} processes started")

    _complete_build.short_description = "Complete build"

    def _cancel_build(self, request, queryset):
        for row in queryset:
            cancel_build.send(row.pk)
        self.message_user(request, f"{len(queryset)} processes started")

    _cancel_build.short_description = "Cancel build"

    def shortfalls(self, obj):
        return obj.shortfalls.count()



import requests
import pytest

from django_ctb.mouser.client import MouserClient


class FakeResponse:
    def __init__(self, text, status_code=200):
        self.text = text
        self.status_code = status_code


class TestMouserClient:
    def test_get_part_missing(self, monkeypatch):
        def fake_post(*args, **kwargs):
            return FakeResponse(
                text="""{
  "Errors": [],
  "SearchResults": {
    "NumberOfResult": 0,
    "Parts": []
  }
} """
            )

        monkeypatch.setattr(requests, "post", fake_post)
        with pytest.raises(MouserClient.EmptyResponse):
            mouser_part = MouserClient().get_part("876-ASDFQWERZXCV")

    def test_get_part_bad(self, monkeypatch):
        def fake_post(*args, **kwargs):
            return FakeResponse(text="bad", status_code=300)

        monkeypatch.setattr(requests, "post", fake_post)
        with pytest.raises(MouserClient.BadResponse):
            mouser_part = MouserClient().get_part("876-ASDFQWERZXCV")

    def test_get_part(self, monkeypatch):
        def fake_post(*args, **kwargs):
            return FakeResponse(
                """{
  "Errors": [],
  "SearchResults": {
    "NumberOfResult": 1,
    "Parts": [
      {
        "Availability": "1498379 In Stock",
        "DataSheetUrl": "https://www.mouser.com/datasheet/2/308/BAT54SLT1_D-2309936.pdf",
        "Description": "Schottky Diodes & Rectifiers 30V 225mW Dual",
        "FactoryStock": "0",
        "ImagePath": "https://www.mouser.com/images/mouserelectronics/images/SOT_23_3_t.jpg",
        "Category": "Schottky Diodes & Rectifiers",
        "LeadTime": "63 Days",
        "LifecycleStatus": null,
        "Manufacturer": "onsemi",
        "ManufacturerPartNumber": "BAT54SLT1G",
        "Min": "1",
        "Mult": "1",
        "MouserPartNumber": "863-BAT54SLT1G",
        "ProductAttributes": [
          {
            "AttributeName": "Packaging",
            "AttributeValue": "Reel"
          },
          {
            "AttributeName": "Packaging",
            "AttributeValue": "Cut Tape"
          },
          {
            "AttributeName": "Packaging",
            "AttributeValue": "MouseReel",
            "AttributeCost": "$7.00 MouseReel™ blah blah blah"
          },
          {
            "AttributeName": "Standard Pack Qty",
            "AttributeValue": "3000"
          }
        ],
        "PriceBreaks": [
          {
            "Quantity": 1,
            "Price": "$0.15",
            "Currency": "USD"
          },
          {
            "Quantity": 10,
            "Price": "$0.106",
            "Currency": "USD"
          },
          {
            "Quantity": 100,
            "Price": "$0.057",
            "Currency": "USD"
          },
          {
            "Quantity": 1000,
            "Price": "$0.031",
            "Currency": "USD"
          },
          {
            "Quantity": 3000,
            "Price": "$0.026",
            "Currency": "USD"
          },
          {
            "Quantity": 9000,
            "Price": "$0.021",
            "Currency": "USD"
          },
          {
            "Quantity": 24000,
            "Price": "$0.019",
            "Currency": "USD"
          },
          {
            "Quantity": 45000,
            "Price": "$0.016",
            "Currency": "USD"
          },
          {
            "Quantity": 99000,
            "Price": "$0.015",
            "Currency": "USD"
          }
        ],
        "AlternatePackagings": null,
        "ProductDetailUrl": "https://www.mouser.com/ProductDetail/onsemi/BAT54SLT1G?qs=vLkC5FC1VN9oCh8qaBIZiQ%3D%3D",
        "Reeling": true,
        "ROHSStatus": "RoHS Compliant",
        "SuggestedReplacement": "",
        "MultiSimBlue": 0,
        "UnitWeightKg": {
          "UnitWeight": 0.00003
        },
        "AvailabilityInStock": "1498379",
        "AvailabilityOnOrder": [],
        "InfoMessages": [],
        "ProductCompliance": [
          {
            "ComplianceName": "USHTS",
            "ComplianceValue": "8541100070"
          },
          {
            "ComplianceName": "CNHTS",
            "ComplianceValue": "8541590000"
          },
          {
            "ComplianceName": "CAHTS",
            "ComplianceValue": "8541100090"
          },
          {
            "ComplianceName": "KRHTS",
            "ComplianceValue": "8541109000"
          },
          {
            "ComplianceName": "TARIC",
            "ComplianceValue": "8541100000"
          },
          {
            "ComplianceName": "MXHTS",
            "ComplianceValue": "8541100101"
          },
          {
            "ComplianceName": "ECCN",
            "ComplianceValue": "EAR99"
          }
        ]
      }
    ]
  }
}"""
            )

        monkeypatch.setattr(requests, "post", fake_post)
        mouser_part = MouserClient().get_part("863-BAT54SLT1G")
        assert mouser_part.name == "BAT54SLT1G"
        assert mouser_part.description == "Schottky Diodes & Rectifiers 30V 225mW Dual"
        assert (
            mouser_part.url_path
            == "/ProductDetail/onsemi/BAT54SLT1G?qs=vLkC5FC1VN9oCh8qaBIZiQ%3D%3D"
        )

    def test_get_part_many_returned(self, monkeypatch):
        def fake_post(*args, **kwargs):
            return FakeResponse(
                """{
  "Errors": [],
  "SearchResults": {
    "NumberOfResult": 2,
    "Parts": [
      {
        "Availability": "1498379 In Stock",
        "DataSheetUrl": "https://www.mouser.com/datasheet/2/308/BAT54SLT1_D-2309936.pdf",
        "Description": "Schottky Diodes & Rectifiers 30V 225mW Dual",
        "FactoryStock": "0",
        "ImagePath": "https://www.mouser.com/images/mouserelectronics/images/SOT_23_3_t.jpg",
        "Category": "Schottky Diodes & Rectifiers",
        "LeadTime": "63 Days",
        "LifecycleStatus": null,
        "Manufacturer": "onsemi",
        "ManufacturerPartNumber": "BAT54SLT1G",
        "Min": "1",
        "Mult": "1",
        "MouserPartNumber": "863-BAT54SLT1G",
        "ProductAttributes": [
          {
            "AttributeName": "Packaging",
            "AttributeValue": "Reel"
          },
          {
            "AttributeName": "Packaging",
            "AttributeValue": "Cut Tape"
          },
          {
            "AttributeName": "Packaging",
            "AttributeValue": "MouseReel",
            "AttributeCost": "$7.00 MouseReel™ blah blah blah"
          },
          {
            "AttributeName": "Standard Pack Qty",
            "AttributeValue": "3000"
          }
        ],
        "PriceBreaks": [
          {
            "Quantity": 1,
            "Price": "$0.15",
            "Currency": "USD"
          },
          {
            "Quantity": 10,
            "Price": "$0.106",
            "Currency": "USD"
          },
          {
            "Quantity": 100,
            "Price": "$0.057",
            "Currency": "USD"
          },
          {
            "Quantity": 1000,
            "Price": "$0.031",
            "Currency": "USD"
          },
          {
            "Quantity": 3000,
            "Price": "$0.026",
            "Currency": "USD"
          },
          {
            "Quantity": 9000,
            "Price": "$0.021",
            "Currency": "USD"
          },
          {
            "Quantity": 24000,
            "Price": "$0.019",
            "Currency": "USD"
          },
          {
            "Quantity": 45000,
            "Price": "$0.016",
            "Currency": "USD"
          },
          {
            "Quantity": 99000,
            "Price": "$0.015",
            "Currency": "USD"
          }
        ],
        "AlternatePackagings": null,
        "ProductDetailUrl": "https://www.mouser.com/ProductDetail/onsemi/BAT54SLT1G?qs=vLkC5FC1VN9oCh8qaBIZiQ%3D%3D",
        "Reeling": true,
        "ROHSStatus": "RoHS Compliant",
        "SuggestedReplacement": "",
        "MultiSimBlue": 0,
        "UnitWeightKg": {
          "UnitWeight": 0.00003
        },
        "AvailabilityInStock": "1498379",
        "AvailabilityOnOrder": [],
        "InfoMessages": [],
        "ProductCompliance": [
          {
            "ComplianceName": "USHTS",
            "ComplianceValue": "8541100070"
          },
          {
            "ComplianceName": "CNHTS",
            "ComplianceValue": "8541590000"
          },
          {
            "ComplianceName": "CAHTS",
            "ComplianceValue": "8541100090"
          },
          {
            "ComplianceName": "KRHTS",
            "ComplianceValue": "8541109000"
          },
          {
            "ComplianceName": "TARIC",
            "ComplianceValue": "8541100000"
          },
          {
            "ComplianceName": "MXHTS",
            "ComplianceValue": "8541100101"
          },
          {
            "ComplianceName": "ECCN",
            "ComplianceValue": "EAR99"
          }
        ]
      },
      {
        "Availability": "1498379 In Stock",
        "DataSheetUrl": "https://www.mouser.com/datasheet/2/308/BAT54SLT1_D-2309936.pdf",
        "Description": "Schottky Diodes & Rectifiers 30V 225mW Dual",
        "FactoryStock": "0",
        "ImagePath": "https://www.mouser.com/images/mouserelectronics/images/SOT_23_3_t.jpg",
        "Category": "Schottky Diodes & Rectifiers",
        "LeadTime": "63 Days",
        "LifecycleStatus": null,
        "Manufacturer": "onsemi",
        "ManufacturerPartNumber": "BAT54SLT1G",
        "Min": "1",
        "Mult": "1",
        "MouserPartNumber": "863-BAT54SLT1G-2",
        "ProductAttributes": [
          {
            "AttributeName": "Packaging",
            "AttributeValue": "Reel"
          },
          {
            "AttributeName": "Packaging",
            "AttributeValue": "Cut Tape"
          },
          {
            "AttributeName": "Packaging",
            "AttributeValue": "MouseReel",
            "AttributeCost": "$7.00 MouseReel™ blah blah blah"
          },
          {
            "AttributeName": "Standard Pack Qty",
            "AttributeValue": "3000"
          }
        ],
        "PriceBreaks": [
          {
            "Quantity": 1,
            "Price": "$0.15",
            "Currency": "USD"
          },
          {
            "Quantity": 10,
            "Price": "$0.106",
            "Currency": "USD"
          },
          {
            "Quantity": 100,
            "Price": "$0.057",
            "Currency": "USD"
          },
          {
            "Quantity": 1000,
            "Price": "$0.031",
            "Currency": "USD"
          },
          {
            "Quantity": 3000,
            "Price": "$0.026",
            "Currency": "USD"
          },
          {
            "Quantity": 9000,
            "Price": "$0.021",
            "Currency": "USD"
          },
          {
            "Quantity": 24000,
            "Price": "$0.019",
            "Currency": "USD"
          },
          {
            "Quantity": 45000,
            "Price": "$0.016",
            "Currency": "USD"
          },
          {
            "Quantity": 99000,
            "Price": "$0.015",
            "Currency": "USD"
          }
        ],
        "AlternatePackagings": null,
        "ProductDetailUrl": "https://www.mouser.com/ProductDetail/onsemi/BAT54SLT1G?qs=vLkC5FC1VN9oCh8qaBIZiQ%3D%3D",
        "Reeling": true,
        "ROHSStatus": "RoHS Compliant",
        "SuggestedReplacement": "",
        "MultiSimBlue": 0,
        "UnitWeightKg": {
          "UnitWeight": 0.00003
        },
        "AvailabilityInStock": "1498379",
        "AvailabilityOnOrder": [],
        "InfoMessages": [],
        "ProductCompliance": [
          {
            "ComplianceName": "USHTS",
            "ComplianceValue": "8541100070"
          },
          {
            "ComplianceName": "CNHTS",
            "ComplianceValue": "8541590000"
          },
          {
            "ComplianceName": "CAHTS",
            "ComplianceValue": "8541100090"
          },
          {
            "ComplianceName": "KRHTS",
            "ComplianceValue": "8541109000"
          },
          {
            "ComplianceName": "TARIC",
            "ComplianceValue": "8541100000"
          },
          {
            "ComplianceName": "MXHTS",
            "ComplianceValue": "8541100101"
          },
          {
            "ComplianceName": "ECCN",
            "ComplianceValue": "EAR99"
          }
        ]
      }
    ]
  }
}"""
            )

        monkeypatch.setattr(requests, "post", fake_post)
        mouser_part = MouserClient().get_part("863-BAT54SLT1G-2")
        assert mouser_part.name == "BAT54SLT1G"
        assert mouser_part.description == "Schottky Diodes & Rectifiers 30V 225mW Dual"
        assert (
            mouser_part.url_path
            == "/ProductDetail/onsemi/BAT54SLT1G?qs=vLkC5FC1VN9oCh8qaBIZiQ%3D%3D"
        )



import requests
import logging

from pydantic import BaseModel, Field, field_validator, ConfigDict

from django.conf import settings

logger = logging.getLogger(__name__)


class MouserPricebreak(BaseModel):
    volume: int = Field(alias="Quantity")
    cost: float = Field(alias="Price")

    model_config = ConfigDict(
        populate_by_name=True,
    )

    @field_validator("cost", mode="before")
    @classmethod
    def remove_prefix(cls, value: str) -> float:
        return float(value.replace("$", ""))


class MouserPart(BaseModel):
    description: str = Field(alias="Description")
    name: str = Field(alias="ManufacturerPartNumber")  # map to value and name
    price_breaks: list[MouserPricebreak] = Field(alias="PriceBreaks")
    url_path: str = Field(alias="ProductDetailUrl")
    mouser_part_number: str = Field(alias="MouserPartNumber")

    model_config = ConfigDict(
        populate_by_name=True,
    )

    @field_validator("url_path", mode="before")
    @classmethod
    def remove_prefix(cls, value: str) -> str:
        return value.replace("https://www.mouser.com", "")


class MouserSearchResponse(BaseModel):
    number_of_result: int = Field(alias="NumberOfResult")
    parts: list[MouserPart] = Field(alias="Parts")


class MouserSearchResponseRoot(BaseModel):
    search_results: MouserSearchResponse = Field(alias="SearchResults")


class MouserSearchByPartRequest(BaseModel):
    mouser_part_number: str = Field(alias="mouserPartNumber")

    model_config = ConfigDict(
        populate_by_name=True,
    )


class MouserSearchByPartRequestRoot(BaseModel):
    search_by_part_request: MouserSearchByPartRequest = Field(
        alias="SearchByPartRequest"
    )

    model_config = ConfigDict(
        populate_by_name=True,
    )


class MouserClient:
    def __init__(self):
        self.api_key = settings.MOUSER_API_KEY

    class BadResponse(Exception):
        pass

    class EmptyResponse(Exception):
        pass

    def get_part(self, mouser_part_number: str) -> MouserPart:
        part_request = MouserSearchByPartRequestRoot(
            search_by_part_request=MouserSearchByPartRequest(
                mouser_part_number=mouser_part_number
            )
        )
        _data = part_request.model_dump_json(by_alias=True)
        # logger.warning(f"posting to get data {_data}")
        response = requests.post(
            "https://api.mouser.com/api/v1/search/partnumber",
            data=_data,
            params={"apiKey": self.api_key},
            headers={"accept": "application/json", "content-type": "application/json"},
        )
        if response.status_code >= 300:
            logger.error(
                f"Part response status code: {response.status_code}: {response.text}"
            )
            raise self.BadResponse
        # logger.warning(f"Part found! {response.text}")
        try:
            response_model = MouserSearchResponseRoot.model_validate_json(response.text)
        except Exception:
            logger.error(f"Can't validate data: {response.text}")
            raise
        if response_model.search_results.number_of_result > 1:
            for part in response_model.search_results.parts:
                if part.mouser_part_number == mouser_part_number:
                    return part
        elif response_model.search_results.number_of_result != 1:
            print(response.text)
            raise self.EmptyResponse
        return response_model.search_results.parts[0]



import dramatiq

from django_ctb import models
from django_ctb.mouser.services import (
    populate_mouser_vendor_part,
)
from django_ctb.services import (
    ProjectVersionBomService,
    ProjectBuildService,
    VendorOrderService,
)


@dramatiq.actor
def sync_project_version(project_version_pk):
    ProjectVersionBomService().sync(project_version_pk)


@dramatiq.actor
def clear_to_build(project_build_pk):
    ProjectBuildService().clear_to_build(project_build_pk)


@dramatiq.actor
def complete_build(project_build_pk):
    ProjectBuildService().complete_build(project_build_pk)


@dramatiq.actor
def complete_order(vendor_order_pk):
    VendorOrderService().complete_order(vendor_order_pk)


@dramatiq.actor
def cancel_build(project_build_pk):
    ProjectBuildService().cancel_build(project_build_pk)



import dramatiq

from django_ctb.mouser.client import MouserClient

from django_ctb import models


class MouserService:
    def _get_price_break(self, price_breaks):
        volume = None
        cost = None
        for price_break in price_breaks:
            if volume and volume >= 10:
                break
            volume = price_break.volume
            cost = price_break.cost
        return volume, cost

    def _populate(self, vendor_part):
        mouser_part = MouserClient().get_part(vendor_part.item_number)
        vendor_part.url_path = mouser_part.url_path
        # find good price
        volume, cost = self._get_price_break(mouser_part.price_breaks)
        vendor_part.cost = cost
        vendor_part.volume = volume
        vendor_part.save()
        vendor_part.part.name = mouser_part.name
        vendor_part.part.value = mouser_part.name
        vendor_part.part.description = mouser_part.description
        vendor_part.part.save()

    class MissingPart(Exception):
        pass

    def _create(self, item_number: str) -> models.VendorPart:
        try:
            mouser_part = MouserClient().get_part(item_number)
        except MouserClient.EmptyResponse:
            raise self.MissingPart
        part, _ = models.Part.objects.get_or_create(
            name=mouser_part.name,
            value=mouser_part.name,
            description=mouser_part.description,
        )
        mouser = models.Vendor.objects.get(name="Mouser")
        volume, cost = self._get_price_break(mouser_part.price_breaks)
        vendor_part, _ = models.VendorPart.objects.update_or_create(
            vendor=mouser,
            part=part,
            item_number=item_number,
            volume=volume,
            cost=cost,
            url_path=mouser_part.url_path,
        )
        return vendor_part

    def populate(self, vendor_part_pk: int):
        try:
            vendor_part = models.VendorPart.objects.get(pk=vendor_part_pk)
        except models.VendorPart.DoesNotExist:
            print(f"No vendor part like {vendor_part_pk} found")
            return
        return self._populate(vendor_part)


@dramatiq.actor
def populate_mouser_vendor_part(vendor_part_pk):
    MouserService().populate(vendor_part_pk)


class MouserPartService:
    """Get data from the Mouser API to create parts"""

    def _get_footprint(self, row):
        footprint, _ = models.Footprint.objects.get_or_create(name=row.footprint_name)
        return footprint

    def _get_package(self, row):
        footprint = self._get_footprint(row)
        try:
            package = models.Package.objects.get(footprints=footprint)
        except models.Package.DoesNotExist:
            _, package_name = row.footprint_name.split(":", maxsplit=1)
            package = models.Package.objects.create(
                technology=models.CircuitTechnologyEnum.UNKNOWN,
                name=package_name,
            )
            package.footprints.add(footprint)
        return package

    def _get_part(self, row):
        # I'm waffling about this cascade setup where each method calls
        #   another to get is prereqs. The prereqs could be passed in as
        #   an alternative, but that means that each method must take in a
        #   snowflake parameter (part, package, footprint)
        package = self._get_package(row)
        assert len(row.symbols) == 1
        symbol = list(row.symbols)[0]
        try:
            part = models.Part.objects.get(
                value=row.value, package=package, symbol=symbol
            )
        except models.Part.DoesNotExist:
            part = models.Part.objects.create(
                name="placeholder",
                value=row.value,
                package=package,
                symbol=symbol,
            )
        return part

    def _get_vendor(self):
        return models.Vendor.objects.get(name="Mouser")

    def create_vendor_part(self, row):
        part = self._get_part(row)
        # create placeholder vendor part
        vendor_part = models.VendorPart.objects.create(
            vendor=self._get_vendor(),
            part=part,
            item_number=row.item_number,
            cost=0.01,
            volume=1,
            url_path="placeholder",
        )
        # send task for worker to populate part/vendor part with API data
        populate_mouser_vendor_part.send(vendor_part.pk)
        # return placeholder vendor part
        return vendor_part



import requests
import pytest
from django.utils import timezone

from django_ctb import models as m

from django_ctb import services as s
from django_ctb.mouser.services import MouserService


class TestVendorOrderService:
    def test__complete_order_line(
        self, vendor_order_line_factory, vendor_part, inventory
    ):
        assert len(m.InventoryLine.objects.all()) == 0
        order_line = vendor_order_line_factory(vendor_part=vendor_part, quantity=11)
        s.VendorOrderService()._complete_order_line(order_line)
        assert len(m.InventoryLine.objects.all()) == 1
        assert len(m.InventoryAction.objects.all()) == 1
        inventory_line = m.InventoryLine.objects.all()[0]
        assert inventory_line.quantity == order_line.quantity
        assert inventory_line.part == order_line.vendor_part.part
        assert inventory_line.inventory == inventory
        action = m.InventoryAction.objects.all()[0]
        assert action.inventory_line == inventory_line
        assert action.delta == 11
        assert action.order_line == order_line
        assert action.build is None
        action.delete()
        inventory_line.delete()

    def test__complete_order_line_existing_inventory(
        self, vendor_order_line_factory, vendor_part, inventory_line_factory
    ):
        inventory_line = inventory_line_factory(part=vendor_part.part, quantity=3)
        order_line = vendor_order_line_factory(vendor_part=vendor_part, quantity=11)
        s.VendorOrderService()._complete_order_line(order_line)
        inventory_line.refresh_from_db()
        assert inventory_line.quantity == 14
        assert len(m.InventoryAction.objects.all()) == 1
        action = m.InventoryAction.objects.all()[0]
        assert action.inventory_line == inventory_line
        assert action.delta == 11
        assert action.order_line == order_line
        assert action.build is None
        action.delete()

    def test__complete_order(
        self,
        vendor_order_line_factory,
        monkeypatch,
        vendor_part_factory,
        vendor_order,
        part_factory,
    ):
        call_args = []

        def fake_complete_order_line(self, order_line):
            call_args.append(order_line)

        order_lines = []
        for idx in range(5):
            part = part_factory(name=f"part{idx}", symbol="R")
            vendor_part = vendor_part_factory(part=part, item_number=f"test-item-{idx}")
            order_lines.append(
                vendor_order_line_factory(vendor_part=vendor_part, quantity=100)
            )

        monkeypatch.setattr(
            s.VendorOrderService, "_complete_order_line", fake_complete_order_line
        )
        s.VendorOrderService()._complete_order(vendor_order)
        vendor_order.refresh_from_db()
        assert vendor_order.fulfilled is not None
        assert call_args == order_lines

    def test_complete_order(self, monkeypatch, vendor_order):
        call_count = 0

        def fake_complete_order(self, order):
            assert order == vendor_order
            nonlocal call_count
            call_count += 1

        monkeypatch.setattr(
            s.VendorOrderService, "_complete_order", fake_complete_order
        )
        s.VendorOrderService().complete_order(vendor_order.pk)
        assert call_count == 1

    def test_complete_order_bad(self, db):
        with pytest.raises(m.VendorOrder.DoesNotExist):
            s.VendorOrderService().complete_order(1234)

    def test_complete_order_fulfilled(self, vendor_order):
        vendor_order.fulfilled = timezone.now()
        vendor_order.save()
        with pytest.raises(m.VendorOrder.DoesNotExist):
            s.VendorOrderService().complete_order(vendor_order.pk)


class TestProjectBuildService:
    def test_part_satisfaction_no_inventory(self, part):
        satisfaction = s.PartSatisfaction(part=part, needed=2)
        assert satisfaction.needed == 2
        assert satisfaction.unfulfilled == 2
        assert satisfaction.fulfillments == []

    def test_part_satisfaction_insufficient_inventory(
        self, part, inventory_line_factory
    ):
        _line = inventory_line_factory(part=part, quantity=1)
        satisfaction = s.PartSatisfaction(part=part, needed=2)
        assert satisfaction.needed == 2
        assert satisfaction.unfulfilled == 1
        assert satisfaction.fulfillments[0].inventory_line == _line
        assert satisfaction.fulfillments[0].depletion == 1

    def test_part_satisfaction_sufficient_inventory(self, part, inventory_line_factory):
        _other_line = inventory_line_factory(part=part, quantity=5)
        _line = inventory_line_factory(part=part, quantity=2)
        satisfaction = s.PartSatisfaction(part=part, needed=2)
        assert satisfaction.needed == 2
        assert satisfaction.unfulfilled == 0
        assert satisfaction.fulfillments[0].inventory_line == _line
        assert satisfaction.fulfillments[0].depletion == 2

    def test_part_satisfaction_sufficient_inventory_deprioritized(
        self, part, inventory_line_factory
    ):
        _other_line = inventory_line_factory(part=part, quantity=5)
        _line = inventory_line_factory(part=part, quantity=2, is_deprioritized=True)
        satisfaction = s.PartSatisfaction(part=part, needed=2)
        assert satisfaction.needed == 2
        assert satisfaction.unfulfilled == 0
        assert satisfaction.fulfillments[0].inventory_line == _other_line
        assert satisfaction.fulfillments[0].depletion == 2

    def test_clear_to_build(self, build, monkeypatch):
        call_count = 0

        def fake_clear_to_build(self, _build):
            assert _build == build
            nonlocal call_count
            call_count += 1

        monkeypatch.setattr(
            s.ProjectBuildService, "_clear_to_build", fake_clear_to_build
        )
        s.ProjectBuildService().clear_to_build(build.pk)
        assert call_count == 1

    def test_clear_to_build_no(self, build):
        build.completed = timezone.now()
        build.save()

        with pytest.raises(m.ProjectBuild.DoesNotExist):
            s.ProjectBuildService().clear_to_build(build.pk)

    def test__clear_to_build(self, project_part, build, inventory_line_factory):
        _line = inventory_line_factory(part=project_part.part, quantity=10)
        s.ProjectBuildService()._clear_to_build(build)
        assert build.part_reservations.count() == 1
        assert len(m.InventoryAction.objects.all()) == 1
        action = m.InventoryAction.objects.all()[0]
        assert action.inventory_line == _line
        assert action.delta == -6
        assert action.order_line is None
        assert action.build == build
        _line.refresh_from_db()
        assert _line.quantity == 4
        s.ProjectBuildPartReservationService().delete_reservations(
            build.part_reservations.all()
        )
        _line.refresh_from_db()
        assert _line.quantity == 10

    def test__clear_to_build_not(
        self, project_part, vendor_part, build, inventory_line_factory
    ):
        _line = inventory_line_factory(part=project_part.part, quantity=1)

        with pytest.raises(s.ProjectBuildService.InsufficientInventory):
            ret = s.ProjectBuildService()._clear_to_build(build)
        assert build.shortfalls.all().count() == 1
        assert build.shortfalls.all()[0].quantity == 5
        assert build.shortfalls.all()[0].part == project_part.part

    def test__clear_to_build_accumulates_by_part(
        self, build, project_part_factory, inventory_line_factory, part
    ):
        _line = inventory_line_factory(part=part, quantity=10)
        reused_project_part = project_part_factory(
            project_version=build.project_version, part=part, quantity=1, line_number=2
        )
        reservations = s.ProjectBuildService()._clear_to_build(build)
        assert len(reservations) == 1
        assert reservations[0].inventory_action.delta == -9
        _line.refresh_from_db()
        assert _line.quantity == 1
        s.ProjectBuildPartReservationService().delete_reservations(
            build.part_reservations.all()
        )
        _line.refresh_from_db()
        assert _line.quantity == 10

    def test__complete_build(self, project_part, build, inventory_line_factory):
        _line = inventory_line_factory(part=project_part.part, quantity=10)
        build.cleared = timezone.now()
        build.save()
        reservation = m.ProjectBuildPartReservation.objects.create(
            inventory_action=None,
            project_build=build,
        )
        s.ProjectBuildService()._complete_build(build)
        reservation.refresh_from_db()
        build.refresh_from_db()
        assert reservation.utilized is not None
        assert build.completed is not None
        reservation.delete()

    def test__complete_build_no(
        self, project_part, build, inventory_line_factory, monkeypatch
    ):
        _line = inventory_line_factory(part=project_part.part, quantity=2)

        call_count = 0

        def fake_clear_to_build(self, _build):
            assert _build == build
            nonlocal call_count
            call_count += 1
            raise s.ProjectBuildService.InsufficientInventory(lacking=[])

        monkeypatch.setattr(
            s.ProjectBuildService, "_clear_to_build", fake_clear_to_build
        )
        with pytest.raises(s.ProjectBuildService.InsufficientInventory):
            ret = s.ProjectBuildService()._complete_build(build)
        build.refresh_from_db()
        assert build.completed is None
        assert call_count == 1

    def test__complete_build_already_completed(self, build, monkeypatch):
        build.completed = timezone.now()
        build.save()

        call_count = 0

        def fake_clear_to_build(self, _build):
            assert _build == build
            nonlocal call_count
            call_count += 1
            raise s.ProjectBuildService.InsufficientInventory(lacking=[])

        monkeypatch.setattr(
            s.ProjectBuildService, "_clear_to_build", fake_clear_to_build
        )

        s.ProjectBuildService()._complete_build(build)
        assert call_count == 0

    def test_complete_build(self, build, monkeypatch):
        call_count = 0
        print(build)
        build.cleared = timezone.now()
        build.save()

        def fake_complete_build(self, _build):
            assert _build == build
            nonlocal call_count
            call_count += 1

        monkeypatch.setattr(
            s.ProjectBuildService, "_complete_build", fake_complete_build
        )
        s.ProjectBuildService().complete_build(build.pk)
        assert call_count == 1

    def test_complete_build_bad(self, build, monkeypatch):
        with pytest.raises(m.ProjectBuild.DoesNotExist):
            s.ProjectBuildService().complete_build(1234)


class TestBillOfMaterialsRow:
    def test_normalize_value(self):
        row = s.BillOfMaterialsRow(
            **{
                "#": 1,
                "Reference": "T1, T2",
                "Qty": 2,
                "PartNum": "ASDF-1234",
                "Vendor": "Mouser",
                "Value": "3M3",
                "Footprint": "Test Footprint",
            }
        )
        assert row.value == "3.3M"

    def test_normalize_value_diode(self):
        row = s.BillOfMaterialsRow(
            **{
                "#": 1,
                "Reference": "T1, T2",
                "Qty": 2,
                "PartNum": "ASDF-1234",
                "Vendor": "Mouser",
                "Value": "1N4148",
                "Footprint": "Test Footprint",
            }
        )
        assert row.value == "1N4148"

    def test_lower_case_si_prefix(self):
        row = s.BillOfMaterialsRow(
            **{
                "#": 1,
                "Reference": "T1, T2",
                "Qty": 2,
                "PartNum": "ASDF-1234",
                "Vendor": "Mouser",
                "Value": "3U3",
                "Footprint": "Test Footprint",
            }
        )
        assert row.value == "3.3u"


class TestMouserPartService:
    @pytest.fixture
    def bom_row(self):
        row = s.BillOfMaterialsRow(
            **{
                "#": 1,
                "Reference": "T1, T2",
                "Qty": 2,
                "PartNum": "ASDF-1234",
                "Vendor": "Mouser",
                "Value": "great",
                "Footprint": "Test Footprint",
            }
        )
        return row

    def test__get_footprint(self, footprint, bom_row):
        _footprint = s.MouserPartService()._get_footprint(bom_row)
        assert footprint == _footprint

    def test__get_footprint_new(self, footprint, bom_row):
        bom_row.footprint_name = "Other Footprint"
        _footprint = s.MouserPartService()._get_footprint(bom_row)
        assert _footprint.name == "Other Footprint"
        assert _footprint != footprint

    def test__get_package(self, package, bom_row):
        _package = s.MouserPartService()._get_package(bom_row)
        assert _package == package

    def test__get_package_new(self, package, bom_row):
        bom_row.footprint_name = "ASDF:Other Footprint"
        _package = s.MouserPartService()._get_package(bom_row)
        assert _package != package
        assert _package.technology == m.CircuitTechnologyEnum.UNKNOWN
        assert _package.name == "Other Footprint"

    def test__get_part(self, part_factory, bom_row):
        part = part_factory(name="Test Part", value="great", symbol="T")
        _part = s.MouserPartService()._get_part(bom_row)
        assert _part == part

    def test__get_part_new(self, part_factory, bom_row, package):
        part = part_factory(name="Test Part", value="awful", symbol="T")
        _part = s.MouserPartService()._get_part(bom_row)
        assert _part != part
        assert _part.name == "placeholder"
        assert _part.value == "great"
        assert _part.package == package
        assert _part.symbol == "T"
        _part.delete()

    def test__get_vendor(self, vendor_mouser):
        _mouser = s.MouserPartService()._get_vendor()
        assert _mouser == vendor_mouser

    def test__get_vendor_missing(self, db):
        with pytest.raises(m.Vendor.DoesNotExist):
            s.MouserPartService()._get_vendor()

    def test_create_vendor_part(
        self, bom_row, part_factory, vendor_mouser, broker, worker, monkeypatch
    ):
        call_count = 0

        def fake_populate(*args, **kwargs):
            nonlocal call_count
            call_count += 1

        monkeypatch.setattr(MouserService, "populate", fake_populate)
        part = part_factory(name="Test Part", value="great", symbol="T")
        # This method should _only_ be hit after a part has been found to not exist
        _vendor_part = s.MouserPartService().create_vendor_part(bom_row)
        assert _vendor_part.part == part
        assert _vendor_part.item_number == "ASDF-1234"
        assert _vendor_part.url_path == "placeholder"

        broker.join("default")
        worker.join()
        assert call_count == 1
        _vendor_part.delete()


class TestProjectVersionBomService:
    def test_get_vendor_part(self, vendor_part):
        _vendor_part = s.ProjectVersionBomService()._get_vendor_part(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "D1, D2",
                    "Qty": 2,
                    "PartNum": vendor_part.item_number,
                    "Vendor": vendor_part.vendor.name,
                    "Value": "LED",
                    "Footprint": "asdf6789",
                }
            ),
        )
        assert _vendor_part == vendor_part

    def test_get_vendor_part_missing(self, vendor_part):
        _vendor_part = s.ProjectVersionBomService()._get_vendor_part(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "D1, D2",
                    "Qty": 2,
                    "PartNum": vendor_part.item_number,
                    "Vendor": "nothing",
                    "Value": "LED",
                    "Footprint": "asdf6789",
                }
            ),
        )
        assert _vendor_part is None

    def test_get_vendor_part_missing_mouser(self, vendor_part_mouser, monkeypatch):
        def fake_create_vendor_part(self, row):
            return vendor_part_mouser

        monkeypatch.setattr(
            s.MouserPartService, "create_vendor_part", fake_create_vendor_part
        )
        _vendor_part = s.ProjectVersionBomService()._get_vendor_part(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "D1, D2",
                    "Qty": 2,
                    "PartNum": "asdf1234",
                    "Vendor": "Mouser",
                    "Value": "LED",
                    "Footprint": "asdf6789",
                }
            ),
        )
        assert _vendor_part == vendor_part_mouser

    def test_get_matching_parts(self, part_factory, footprint):
        green_led = part_factory(name="LED Green", value="LED", symbol="D")
        white_led = part_factory(name="LED White", value="LED", symbol="D")
        parts = s.ProjectVersionBomService()._get_matching_parts(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "D1, D2",
                    "Qty": 2,
                    "PartNum": "asdf1234",
                    "Vendor": "vendor",
                    "Value": "LED",
                    "Footprint": footprint.name,
                }
            ),
        )
        assert green_led in parts
        assert white_led in parts

    def test_get_matching_parts_discriminating(self, part_factory, footprint):
        log_pot = part_factory(name="Spinny Boi Pot", value="A100K", symbol="RV")
        lin_pot = part_factory(name="Spinny Boi Pot", value="B100K", symbol="RV")
        parts = s.ProjectVersionBomService()._get_matching_parts(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "RV1, RV2",
                    "Qty": 2,
                    "PartNum": "asdf1234",
                    "Vendor": "vendor",
                    "Value": "B100K",
                    "Footprint": footprint.name,
                }
            ),
        )
        assert lin_pot in parts
        assert log_pot not in parts

    def test_get_matching_parts_quantity_sorting(
        self, part_factory, footprint, inventory_line_factory
    ):
        green_led = part_factory(name="LED Green", value="LED", symbol="D")
        purple_led = part_factory(name="LED purple", value="LED", symbol="D")
        white_led = part_factory(name="LED White", value="LED", symbol="D")
        inventory_line_factory(part=green_led, quantity=53)
        inventory_line_factory(part=white_led, quantity=47)
        inventory_line_factory(part=purple_led, quantity=22)
        parts = s.ProjectVersionBomService()._get_matching_parts(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "D1, D2",
                    "Qty": 2,
                    "PartNum": "asdf1234",
                    "Vendor": "vendor",
                    "Value": "LED",
                    "Footprint": footprint.name,
                }
            ),
        )
        assert green_led == parts[0]
        assert white_led == parts[1]
        assert purple_led == parts[2]
        assert len(parts) == 3
        for part in parts:
            if part == green_led:
                assert part.qty_in_inventory == 53
            elif part == white_led:
                assert part.qty_in_inventory == 47
            elif part == purple_led:
                assert part.qty_in_inventory == 22

    def test_get_matching_parts_deprioritized(
        self, part_factory, footprint, inventory_line_factory
    ):
        green_led = part_factory(name="LED Green", value="LED", symbol="D")
        white_led = part_factory(name="LED white", value="LED", symbol="D")
        inventory_line_factory(part=green_led, quantity=53)
        inventory_line_factory(part=white_led, quantity=22, is_deprioritized=True)
        parts = s.ProjectVersionBomService()._get_matching_parts(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "D1, D2",
                    "Qty": 2,
                    "PartNum": "asdf1234",
                    "Vendor": "vendor",
                    "Value": "LED",
                    "Footprint": footprint.name,
                }
            ),
        )
        assert white_led not in parts

    def test_get_part_vendor_part(self, monkeypatch, vendor_part):
        call_count = 0

        def fake_get_vendor_part(self, *, row):
            assert row.vendor_name == "vendor"
            assert row.item_number == "asdf1234"
            nonlocal call_count
            call_count += 1
            return vendor_part

        monkeypatch.setattr(
            s.ProjectVersionBomService, "_get_vendor_part", fake_get_vendor_part
        )
        part = s.ProjectVersionBomService()._get_part(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "F1, F2",
                    "Qty": 2,
                    "PartNum": "asdf1234",
                    "Vendor": "vendor",
                    "Value": "asdf6789",
                    "Footprint": "asdf1234",
                }
            ),
        )
        assert call_count == 1
        assert part == vendor_part.part

    def test_get_part_regular(self, monkeypatch, part_queryset, part):
        call_count = 0

        def fake_get_matching_part(self, *, row):
            assert row.value == "asdf6789"
            assert row.footprint_name == "asdf1234"
            assert row.symbols == {"F"}
            nonlocal call_count
            call_count += 1
            return part_queryset

        monkeypatch.setattr(
            s.ProjectVersionBomService, "_get_matching_parts", fake_get_matching_part
        )
        _part = s.ProjectVersionBomService()._get_part(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "F1, F2",
                    "Qty": 2,
                    "PartNum": "",
                    "Vendor": "",
                    "Value": "asdf6789",
                    "Footprint": "asdf1234",
                }
            ),
        )
        assert call_count == 1
        assert _part == part

    def test_get_part_missing(self, monkeypatch):
        call_count = 0

        def fake_get_matching_part(self, *, row):
            assert row.value == "asdf6789"
            assert row.footprint_name == "asdf1234"
            assert row.symbols == {"F"}
            nonlocal call_count
            call_count += 1
            return m.Part.objects.none()

        monkeypatch.setattr(
            s.ProjectVersionBomService, "_get_matching_parts", fake_get_matching_part
        )
        _part = s.ProjectVersionBomService()._get_part(
            row=s.BillOfMaterialsRow(
                **{
                    "#": 1,
                    "Reference": "F1, F2",
                    "Qty": 2,
                    "PartNum": "",
                    "Vendor": "",
                    "Value": "asdf6789",
                    "Footprint": "asdf1234",
                }
            ),
        )
        assert call_count == 1
        assert _part is None

    def test_build_bom_url(self, project_version):
        ret = s.ProjectVersionBomService()._build_bom_url(project_version)
        assert ret == "https://gitbub.com/fake/fake/raw/v0/nested/deep/test.csv"

    def test_sync_footprints(self, project_part):
        _old_footprint_ref = m.ProjectPartFootprintRef.objects.create(
            project_part=project_part,
            footprint_ref="F2",
        )
        _old_footprint_ref2 = m.ProjectPartFootprintRef.objects.create(
            project_part=project_part,
            footprint_ref="F3",
        )
        s.ProjectVersionBomService()._sync_footprints(
            {"F1", "F3"}, project_part=project_part
        )
        with pytest.raises(m.ProjectPartFootprintRef.DoesNotExist):
            _old_footprint_ref.refresh_from_db()
        refs = m.ProjectPartFootprintRef.objects.filter(project_part=project_part)
        assert _old_footprint_ref2 in refs
        assert "F1" in refs.values_list("footprint_ref", flat=True)

    def test_sync_implicit_parts(
        self,
        project_version,
        part,
        part_factory,
        project_part_factory,
        implicit_project_part_factory,
    ):
        implicit_part = part_factory(name="implicit part", symbol="IP")
        project_part = project_part_factory(
            project_version=project_version, part=part, line_number=69
        )
        implicit_project_part = implicit_project_part_factory(
            for_package=part.package, part=implicit_part, quantity=3
        )
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 0
        s.ProjectVersionBomService()._sync_implicit_parts(project_part=project_part)
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 1
        pp = m.ProjectPart.objects.filter(is_implicit=True)[0]
        assert pp.quantity == 6
        assert pp.line_number == project_part.line_number
        pp.delete()

    def test_sync_implicit_parts_remove_old(
        self,
        project_version,
        part,
        part_factory,
        project_part_factory,
        implicit_project_part_factory,
    ):
        implicit_part = part_factory(name="implicit part", symbol="IP")
        old_part = part_factory(name="old part", symbol="IP")
        project_part = project_part_factory(
            project_version=project_version, part=part, line_number=69
        )
        old_project_part = project_part_factory(
            project_version=project_version,
            part=old_part,
            line_number=69,
            is_implicit=True,
        )
        implicit_project_part = implicit_project_part_factory(
            for_package=part.package, part=implicit_part, quantity=3
        )
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 1
        s.ProjectVersionBomService()._sync_implicit_parts(project_part=project_part)
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 1
        pp = m.ProjectPart.objects.filter(is_implicit=True)[0]
        assert pp.quantity == 6
        assert pp.line_number == project_part.line_number
        assert pp.part == implicit_part
        with pytest.raises(m.ProjectPart.DoesNotExist):
            old_project_part.refresh_from_db()
        pp.delete()

    def test_sync_implicit_parts_update(
        self,
        project_version,
        part,
        part_factory,
        project_part_factory,
        implicit_project_part_factory,
    ):
        implicit_part = part_factory(name="implicit part", symbol="IP")
        project_part = project_part_factory(
            project_version=project_version, part=part, line_number=69
        )
        old_project_part = project_part_factory(
            project_version=project_version,
            part=implicit_part,
            line_number=69,
            is_implicit=True,
            quantity=9,
        )
        implicit_project_part = implicit_project_part_factory(
            for_package=part.package, part=implicit_part, quantity=3
        )
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 1
        s.ProjectVersionBomService()._sync_implicit_parts(project_part=project_part)
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 1
        pp = m.ProjectPart.objects.filter(is_implicit=True)[0]
        assert pp.quantity == 6
        assert pp.line_number == project_part.line_number
        assert pp.part == implicit_part
        assert pp == old_project_part
        pp.delete()

    def test_sync_implicit_parts_multiple(
        self,
        project_version,
        part,
        part_factory,
        project_part_factory,
        implicit_project_part_factory,
    ):
        implicit_part = part_factory(name="implicit part", symbol="IP")
        other_implicit_part = part_factory(name="other_implicit part", symbol="IP")
        project_part = project_part_factory(
            project_version=project_version, part=part, line_number=69
        )
        old_project_part = project_part_factory(
            project_version=project_version,
            part=implicit_part,
            line_number=69,
            is_implicit=True,
            quantity=9,
        )
        other_old_project_part = project_part_factory(
            project_version=project_version,
            part=other_implicit_part,
            line_number=69,
            is_implicit=True,
            quantity=2,
        )
        implicit_project_part = implicit_project_part_factory(
            for_package=part.package, part=implicit_part, quantity=3
        )
        other_implicit_project_part = implicit_project_part_factory(
            for_package=part.package, part=other_implicit_part, quantity=2
        )
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 2
        s.ProjectVersionBomService()._sync_implicit_parts(project_part=project_part)
        assert m.ProjectPart.objects.filter(is_implicit=True).count() == 2
        pp = m.ProjectPart.objects.filter(is_implicit=True, part=implicit_part)[0]
        assert pp.quantity == 6
        assert pp.line_number == project_part.line_number
        assert pp == old_project_part
        pp.delete()
        opp = m.ProjectPart.objects.filter(is_implicit=True, part=other_implicit_part)[
            0
        ]
        assert opp == other_old_project_part
        assert opp.line_number == project_part.line_number
        assert opp.quantity == 4
        opp.delete()

    def test_sync_row(self, project_version, part, monkeypatch):
        _row = s.BillOfMaterialsRow(
            **{
                "#": 1,
                "Reference": "A1, A2, A33, D12",
                "#": 69,
                "Qty": 420,
                "PartNum": None,
                "Vendor": None,
                "Value": "asdf6789",
                "Footprint": "asdf1234",
            }
        )

        def fake_get_part(self, *, row):
            assert row.symbols == {"A", "D"}
            assert row == _row
            return part

        monkeypatch.setattr(s.ProjectVersionBomService, "_get_part", fake_get_part)
        project_part = s.ProjectVersionBomService()._sync_row(
            row=_row, project_version=project_version
        )
        assert project_part.project_version == project_version
        assert project_part.part == part
        assert project_part.line_number == 69
        assert project_part.quantity == 420
        refs = project_part.footprint_refs.all()
        assert len(refs) == 4
        _refs = {"A1", "A2", "A33", "D12"}
        for ref in refs:
            _refs.remove(ref.footprint_ref)
        assert len(_refs) == 0
        project_part.delete()

    def test__sync(self, project_version, monkeypatch, project_part_factory, part):
        class Closable:
            def __init__(self, content):
                self.content = content

            def close(self):
                pass

        def fake_get(url):
            return Closable(
                b"""Qty,Reference,Vendor,PartNum,Footprint,Value
3,"A1, A2, A3","test vendor","test-item-number","Test Footprint","asdf"
"""
            )

        def fake_bom_url(self, project_version):
            return "hyup"

        _real_project_part = project_part_factory(
            project_version=project_version, part=part
        )
        _bad_project_part = project_part_factory(
            project_version=project_version, part=part
        )

        def fake_sync_row(self, *, row, project_version):
            return _real_project_part

        monkeypatch.setattr(s.ProjectVersionBomService, "_build_bom_url", fake_bom_url)
        monkeypatch.setattr(s.ProjectVersionBomService, "_sync_row", fake_sync_row)
        monkeypatch.setattr(requests, "get", fake_get)

        s.ProjectVersionBomService()._sync(project_version)
        with pytest.raises(m.ProjectPart.DoesNotExist):
            _bad_project_part.refresh_from_db()
        _project_parts = project_version.project_parts.all()
        assert len(_project_parts) == 1
        assert _real_project_part in _project_parts

    def test__sync_missing_part(
        self, project_version, monkeypatch, project_part_factory, part
    ):
        class Closable:
            def __init__(self, content):
                self.content = content

            def close(self):
                pass

        def fake_get(url):
            return Closable(
                b"""#,Qty,Reference,Vendor,PartNum,Footprint,Value
1,3,"A1, A2, A3",,,"Unknown Footprint","zxcv"
"""
            )

        def fake_bom_url(self, project_version):
            return "hyup"

        _bad_project_part = project_part_factory(
            project_version=project_version, part=part, line_number=2
        )
        print(_bad_project_part)
        print(_bad_project_part.line_number)
        # def fake_sync_row(self, *, row, project_version):
        #     raise s.ProjectVersionBomService.PartMissing("blah")

        monkeypatch.setattr(s.ProjectVersionBomService, "_build_bom_url", fake_bom_url)
        # monkeypatch.setattr(s.ProjectVersionBomService, "_sync_row", fake_sync_row)
        monkeypatch.setattr(requests, "get", fake_get)

        ret = s.ProjectVersionBomService()._sync(project_version)
        with pytest.raises(m.ProjectPart.DoesNotExist):
            _bad_project_part.refresh_from_db()
        _project_parts = project_version.project_parts.all()
        assert len(_project_parts) == 1
        assert _project_parts[0].part is None
        assert _project_parts[0].missing_part_description is not None

    def test_sync(self, project_version, monkeypatch):
        call_count = 0

        def fake_sync(self, _project_version):
            nonlocal call_count
            call_count += 1
            assert project_version == _project_version
            return {}

        monkeypatch.setattr(s.ProjectVersionBomService, "_sync", fake_sync)
        s.ProjectVersionBomService().sync(project_version.pk)
        assert call_count == 1



import csv
import re
import io
import logging
import zipfile
from contextlib import closing

from django.db.models import Sum
from django.utils import timezone
from pydantic import BaseModel, Field, field_validator, AliasChoices
import requests

from django_ctb import models
from django_ctb.mouser.services import MouserPartService

logger = logging.getLogger(__name__)


class VendorOrderService:
    def _complete_order_line(self, order_line):
        # resolve part/inventory line
        inventory_line, _ = models.InventoryLine.objects.get_or_create(
            part=order_line.vendor_part.part, inventory=order_line.for_inventory
        )
        # create inventory action
        inventory_action = models.InventoryAction.objects.create(
            inventory_line=inventory_line,
            delta=order_line.quantity,
            order_line=order_line,
        )
        # update inventory line
        inventory_line.quantity += order_line.quantity
        inventory_line.save()

    def _complete_order(self, order):
        for order_line in order.lines.all():
            self._complete_order_line(order_line)
        order.fulfilled = timezone.now()
        order.save()

    def complete_order(self, order_pk):
        try:
            order = (
                models.VendorOrder.objects.filter(fulfilled__isnull=True)
                .prefetch_related("lines")
                .get(pk=order_pk)
            )
        except models.VendorOrder.DoesNotExist:
            raise
        self._complete_order(order)


class InventoryLineFulfillment:
    def __init__(self, *, inventory_line, depletion):
        self.inventory_line = inventory_line
        self.depletion = depletion

    def execute_fulfillment(self):
        self.inventory_line.quantity -= self.depletion
        self.inventory_line.save()


class PartSatisfaction:
    def __init__(self, *, part, needed):
        self.part = part
        self.needed = needed
        self.unfulfilled = needed
        self.fulfillments = self.calculate_fulfillment()

    def calculate_fulfillment(self) -> list[InventoryLineFulfillment]:
        inventory_lines = models.InventoryLine.objects.filter(
            part=self.part, is_deprioritized=False
        ).order_by("quantity")
        needed_lines = []
        for inventory_line in inventory_lines:
            depletion = min(self.unfulfilled, inventory_line.quantity)
            needed_lines.append(
                InventoryLineFulfillment(
                    inventory_line=inventory_line, depletion=depletion
                )
            )
            self.unfulfilled -= depletion
            if self.unfulfilled == 0:
                break
        return needed_lines


class ProjectBuildPartReservationService:
    def create_reservations(
        self, satisfaction: PartSatisfaction, build: models.ProjectBuild
    ) -> list[models.ProjectBuildPartReservation]:
        reservations = []
        for fulfillment in satisfaction.fulfillments:
            # Do inventory action
            inventory_action = models.InventoryAction.objects.create(
                inventory_line=fulfillment.inventory_line,
                delta=fulfillment.depletion * -1,
                build=build,
            )
            fulfillment.execute_fulfillment()

            # create reservations
            reservations.append(
                models.ProjectBuildPartReservation.objects.create(
                    inventory_action=inventory_action,
                    project_build=build,
                )
            )
        return reservations

    def delete_reservations(self, reservations):
        for reservation in reservations:
            # undo inventory action
            inventory_action = reservation.inventory_action
            if inventory_action is not None:
                reservation.inventory_action = None
                reservation.save()
                inventory_action.inventory_line.quantity -= inventory_action.delta
                inventory_action.inventory_line.save()
                inventory_action.delete()
            reservation.delete()


class ProjectBuildService:
    class InsufficientInventory(Exception):
        def __init__(self, *args, lacking, **kwargs):
            self.lacking = lacking
            super().__init__(*args, **kwargs)

    def _clear_to_build(self, build) -> list[models.ProjectBuildPartReservation]:
        if build.cleared is not None:
            print(f"!! Build already cleared at {build.cleared}")
            return list(
                models.ProjectBuildPartReservation.objects.filter(
                    project_build=build, utilized__isnull=True
                )
            )
        consolodated_project_parts: dict[int, int] = {}
        for project_part in build.project_version.project_parts.all():
            _pk = project_part.part.pk
            consolodated_project_parts.setdefault(
                _pk, {"part": project_part.part, "quantity": 0}
            )
            consolodated_project_parts[_pk]["quantity"] += (
                project_part.quantity * build.quantity
            )

        satisfied = []
        unsatisfied_demand = []
        for _part in consolodated_project_parts.values():
            part = _part["part"]
            quantity_needed = _part["quantity"]
            satisfaction = PartSatisfaction(part=part, needed=quantity_needed)
            if satisfaction.unfulfilled > 0:
                unsatisfied_demand.append(satisfaction)
                models.ProjectBuildPartShortage.objects.create(
                    project_build=build,
                    part=part,
                    quantity=satisfaction.unfulfilled,
                )
            else:
                satisfied.append(satisfaction)

        if unsatisfied_demand:
            print("!! Not clear to build !!")
            print("!! Lacking: ")
            for unsatisfied in unsatisfied_demand:
                _vendor_part = (
                    unsatisfied.part.part_vendors.all().order_by("cost").first()
                )
                print(unsatisfied.part, unsatisfied.unfulfilled)
                if _vendor_part:
                    print(f">> {_vendor_part.vendor.name} {_vendor_part.item_number}")
            raise self.InsufficientInventory(lacking=unsatisfied_demand)

        reservations = []
        for satisfaction in satisfied:
            # create reservation for part
            reservations.extend(
                ProjectBuildPartReservationService().create_reservations(
                    satisfaction=satisfaction, build=build
                )
            )
        build.cleared = timezone.now()
        build.save()
        return reservations

    def _complete_build(self, build):
        if build.completed is not None:
            print(f"!! Build already completed at {build.completed}")
            return
        reservations = self._clear_to_build(build)

        for reservation in reservations:
            reservation.utilized = timezone.now()
            reservation.save()

        build.completed = timezone.now()
        build.save()
        return

    def complete_build(self, build_pk):
        try:
            build = (
                models.ProjectBuild.objects.filter(completed__isnull=True)
                .exclude(cleared__isnull=True)
                .select_related("project_version")
                .prefetch_related("project_version__parts")
                .get(pk=build_pk)
            )
        except models.ProjectBuild.DoesNotExist:
            raise
        return self._complete_build(build)

    def clear_to_build(self, build_pk):
        try:
            build = (
                models.ProjectBuild.objects.filter(completed__isnull=True)
                .select_related("project_version")
                .prefetch_related("project_version__parts")
                .get(pk=build_pk)
            )
        except models.ProjectBuild.DoesNotExist:
            raise
        try:
            return self._clear_to_build(build)
        except self.InsufficientInventory:
            return []

    def _cancel_build(self, build):
        if build.completed is not None:
            print(f"!! Build already completed, cannot cancel")
            return
        ProjectBuildPartReservationService().delete_reservations(
            build.part_reservations.all()
        )
        build.cleared = None
        build.save()

    def cancel_build(self, build_pk):
        try:
            build = (
                models.ProjectBuild.objects.filter(completed__isnull=True)
                # .select_related("part_reservations")
                .prefetch_related("part_reservations").get(pk=build_pk)
            )
        except models.ProjectBuild.DoesNotExist:
            raise
        try:
            return self._cancel_build(build)
        except:
            return


class BillOfMaterialsRow(BaseModel):
    line_number: int | None = Field(validation_alias=AliasChoices("#", "line"))
    references: list[str] = Field(validation_alias=AliasChoices("Reference", "Ref"))
    quantity: int = Field(validation_alias=AliasChoices("Qty", "Qnty"))
    value: str = Field(alias="Value")
    footprint_name: str = Field(alias="Footprint")
    vendor_name: str | None = Field(alias="Vendor", default=None)
    item_number: str | None = Field(alias="PartNum", default=None)

    @property
    def symbols(self):
        return {r.strip("0123456789") for r in self.references}

    @field_validator("value", mode="before")
    @classmethod
    def normalize_value(cls, value: str):
        """3M3 -> 3.3M"""
        matches = re.match(r"(?P<whole>\d+)(?P<prefix>[mMkKuUpPn])(?P<frac>\d+)", value)
        if matches is not None:
            _prefix = matches.group("prefix")
            if _prefix in "UP":
                _prefix = _prefix.lower()
            value = f"{matches.group('whole')}.{matches.group('frac')}" f"{_prefix}"
        return value

    @field_validator("references", mode="before")
    @classmethod
    def split_and_strip(cls, value: str) -> list[str]:
        return [v.strip() for v in value.split(",") if v.strip()]


class ProjectVersionBomService:
    """Downloads BOM from repo at specified commit and creates project
    parts for each line."""

    def _get_vendor_part(self, *, row):
        try:
            vendor_part = models.VendorPart.objects.get(
                vendor__name=row.vendor_name, item_number=row.item_number
            )
        except models.VendorPart.DoesNotExist:
            # such a vendor part will need to exist before this project
            #  bom can be validated
            logger.error(
                f"Project BOM includes vendor specified part which is not "
                f"in parts library: {row.vendor_name}: {row.item_number}"
            )
            # If the vendor is Mouser then the part can be looked up via API
            if row.vendor_name == "Mouser":
                vendor_part = MouserPartService().create_vendor_part(
                    row=row,
                )
            else:
                return None
        return vendor_part

    def _get_matching_parts(self, *, row):
        return (
            models.Part.objects.filter(
                value=row.value,
                package__footprints__name=row.footprint_name,
                symbol__in=row.symbols,
            )
            .exclude(inventory_lines__is_deprioritized=True)
            .annotate(qty_in_inventory=Sum("inventory_lines__quantity"))
            .order_by("-qty_in_inventory")
        )

    def _get_part(self, *, row):
        if row.item_number and row.vendor_name:
            return self._get_vendor_part(row=row).part

        # Consider cases where there is more than one part that satisfies,
        #  e.g. an LED (parts may include a white LED and a green LED).
        #  * Don't choose the part which is deprioritized
        #  * Choose the part which has inventory!

        _part = self._get_matching_parts(row=row).first()
        return _part

    def _sync_footprints(serf, footprint_refs, *, project_part):
        # get rid of any outdated/altered refs
        models.ProjectPartFootprintRef.objects.filter(
            project_part=project_part
        ).exclude(footprint_ref__in=footprint_refs).delete()
        # create any missing refs
        for footprint_ref in footprint_refs:
            models.ProjectPartFootprintRef.objects.get_or_create(
                footprint_ref=footprint_ref, project_part=project_part
            )

    def _sync_implicit_parts(self, *, project_part):
        if project_part.part is None:
            print(
                "Cannot create implicit parts for project_part lacking a "
                f"part {project_part}"
            )
            return
        implicit_project_parts = models.ImplicitProjectPart.objects.filter(
            for_package=project_part.part.package
        )
        project_part_pks = []
        for implicit_project_part in implicit_project_parts:
            print(implicit_project_part)
            quantity = implicit_project_part.quantity * project_part.quantity
            _implicit_project_part, _ = models.ProjectPart.objects.update_or_create(
                project_version=project_part.project_version,
                line_number=project_part.line_number,
                is_implicit=True,
                part=implicit_project_part.part,
                defaults={
                    "quantity": quantity,
                },
            )
            print(f"created implicit project part {_implicit_project_part.pk}")
            project_part_pks.append(_implicit_project_part.pk)
        # clean up any vestiges
        models.ProjectPart.objects.exclude(pk__in=project_part_pks).filter(
            line_number=project_part.line_number, is_implicit=True
        ).delete()

    def _sync_row(self, *, row, project_version):
        _part = self._get_part(row=row)
        defaults = {"part": _part, "quantity": row.quantity}
        if _part is None:
            defaults.update({"missing_part_description": f"{row}"})
        project_part, _ = models.ProjectPart.objects.update_or_create(
            project_version=project_version,
            line_number=row.line_number,
            is_implicit=False,
            defaults=defaults,
        )
        self._sync_footprints(row.references, project_part=project_part)
        self._sync_implicit_parts(project_part=project_part)
        return project_part

    def _build_bom_url(self, project_version):
        # build file url
        # e.g. https://github.com/thismatters/EurorackLfo/raw/main/lfo.csv
        _bom_path = project_version.bom_path
        if not _bom_path.startswith("/"):
            _bom_path = "/" + _bom_path
        _bom_file = (
            project_version.project.git_url
            + "/raw/"
            + project_version.commit_ref
            + _bom_path
        )
        return _bom_file

    def _sync(self, project_version):
        row_errors = {}
        project_part_pks = []
        _bom_url = self._build_bom_url(project_version=project_version)
        print(_bom_url)
        file_response = requests.get(_bom_url)
        with closing(file_response), io.StringIO(
            file_response.content.decode("utf-8")
        ) as bom:
            reader = csv.DictReader(bom)
            print("Parsing csv")
            for line_number, _row in enumerate(reader, start=1):
                if "#" not in _row:
                    _row["#"] = line_number
                row = BillOfMaterialsRow(**_row)
                project_part = self._sync_row(row=row, project_version=project_version)
                if project_part.part is None:
                    logger.warning(f"Part missing for row {row}")
                    logger.warning(row.symbols)
                    row_errors.setdefault("part_missing", []).append(row.line_number)
                project_part_pks.append(project_part.pk)
        print(f"created these project parts {project_part_pks}")
        # remove any outdated lines
        print(
            models.ProjectPart.objects.filter(
                project_version=project_version, is_implicit=False
            )
            .exclude(pk__in=project_part_pks)
            .delete()
        )
        project_version.synced = timezone.now()
        project_version.save()
        return row_errors

    def sync(self, project_version_pk):
        project_version = models.ProjectVersion.objects.get(pk=project_version_pk)
        return self._sync(project_version)



from django.db import models
from django.utils import timezone

from django_enumfield import enum


class UnitEnum(enum.Enum):
    NONE = 0
    OHM = 1
    FARAD = 2
    HENRY = 3
    VOLT = 4
    AMPERE = 5

    __default__ = NONE

    __labels__ = {
        NONE: "None",
        OHM: "Ohm",
        FARAD: "F",
        HENRY: "H",
        VOLT: "V",
        AMPERE: "A",
    }


class CircuitTechnologyEnum(enum.Enum):
    THROUGH_HOLE = 0
    SURFACE_MOUNT = 1
    UNKNOWN = 2

    __labels__ = {
        THROUGH_HOLE: "THT",
        SURFACE_MOUNT: "SMD",
    }


class Footprint(models.Model):
    """Names of footprints as they might appear in BOM"""

    name = models.CharField(max_length=64)

    def __str__(self):  # pragma: no cover
        return self.name


class Package(models.Model):
    technology = enum.EnumField(CircuitTechnologyEnum)
    name = models.CharField(max_length=32, help_text="e.g. 0805, or TO-92W")
    footprints = models.ManyToManyField(Footprint, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.technology.label} {self.name}"


class Vendor(models.Model):
    name = models.CharField(max_length=64)
    base_url = models.URLField(help_text="sans trailing slash plz")

    def __str__(self):  # pragma: no cover
        return self.name


class Part(models.Model):
    name = models.CharField(max_length=64)
    description = models.CharField(max_length=256, null=True, blank=True)
    value = models.CharField(
        max_length=32,
        help_text="Value as found on BOM e.g. '10K', '22u'",
        null=True,
        blank=True,
    )
    tolerance = models.SmallIntegerField(
        help_text="as percentage", null=True, blank=True
    )
    loading_limit = models.CharField(max_length=16, null=True, blank=True)
    unit = enum.EnumField(UnitEnum)
    symbol = models.CharField(
        max_length=4,
        help_text="What prefix might this part have on a circuit schematic?",
        null=True,
        blank=True,
    )
    package = models.ForeignKey(
        Package, on_delete=models.PROTECT, null=True, blank=True
    )
    vendors = models.ManyToManyField(Vendor, related_name="parts", through="VendorPart")
    _is_scraped = models.BooleanField(default=False)

    def __str__(self):  # pragma: no cover
        return f"{self.name} {self.symbol} {self.value} -- {self.package}"

    @property
    def unit_cost(self):
        part_vendor = self.part_vendors.all().order_by("cost").first()
        if part_vendor is not None:
            return part_vendor.cost
        return 0


class ImplicitProjectPart(models.Model):
    """Certain parts do not appear on the BOM, but must be used for the final
    build. These are represented here. e.g. LED bezel, potentiometer knob, even
    the PCB could fit in this category?"""

    part = models.ForeignKey(Part, on_delete=models.PROTECT)
    for_package = models.ForeignKey(Package, on_delete=models.PROTECT)
    quantity = models.SmallIntegerField(default=1)


class VendorPart(models.Model):
    vendor = models.ForeignKey(
        Vendor, related_name="vendor_parts", on_delete=models.PROTECT
    )
    part = models.ForeignKey(
        Part, related_name="part_vendors", on_delete=models.CASCADE
    )
    item_number = models.CharField(
        max_length=64, help_text="how the vendor identifies the part"
    )
    cost = models.DecimalField(
        decimal_places=4,
        max_digits=8,
        help_text="Cost per unit when purchased at given volume",
        null=True,
        blank=True,
    )
    volume = models.PositiveIntegerField(
        help_text="Must purchase this many to get cost",
        null=True,
        blank=True,
    )
    url_path = models.CharField(
        max_length=128, help_text="includes the leading slash!", null=True, blank=True
    )
    # unit_cost = models.GeneratedField(
    #     expression=models.F("cost") / models.F("volume"),
    #     output_field=models.DecimalField(decimal_places=4, max_digits=8),
    #     db_persist=False,
    # )

    @property
    def url(self):  # pragma: no cover
        return f"{self.vendor.base_url}{self.url_path}"

    def __str__(self):  # pragma: no cover
        return f"{self.vendor} - {self.item_number}"

    class Meta:
        ordering = ("vendor__name", "item_number")


class VendorOrder(models.Model):
    vendor = models.ForeignKey(Vendor, on_delete=models.PROTECT)
    order_number = models.CharField(max_length=128, null=True)
    created = models.DateTimeField(default=timezone.now)
    fulfilled = models.DateTimeField(null=True, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.vendor} #{self.order_number}"


class VendorOrderLine(models.Model):
    vendor_order = models.ForeignKey(
        VendorOrder, on_delete=models.CASCADE, related_name="lines"
    )
    vendor_part = models.ForeignKey(
        VendorPart, on_delete=models.PROTECT, related_name="lines"
    )
    quantity = models.PositiveIntegerField()
    cost = models.DecimalField(decimal_places=4, max_digits=8, help_text="per unit")
    for_inventory = models.ForeignKey("Inventory", on_delete=models.PROTECT)


class Inventory(models.Model):
    name = models.CharField(max_length=64)

    class Meta:
        verbose_name_plural = "Inventories"

    def __str__(self):  # pragma: no cover
        return self.name


class InventoryLine(models.Model):
    created = models.DateTimeField(default=timezone.now)
    updated = models.DateTimeField(auto_now=True)
    inventory = models.ForeignKey(Inventory, on_delete=models.PROTECT)
    part = models.ForeignKey(
        Part, on_delete=models.PROTECT, related_name="inventory_lines"
    )
    quantity = models.IntegerField(default=0)
    is_deprioritized = models.BooleanField(default=False)


class InventoryAction(models.Model):
    inventory_line = models.ForeignKey(InventoryLine, on_delete=models.CASCADE)
    # what happened
    delta = models.IntegerField()
    # why did it happen (order fulfilled | project built | correction)
    order_line = models.ForeignKey(
        VendorOrderLine, on_delete=models.PROTECT, null=True, blank=True
    )
    build = models.ForeignKey(
        "ProjectBuild", on_delete=models.PROTECT, null=True, blank=True
    )
    # when did it happen
    created = models.DateTimeField(default=timezone.now)


class Project(models.Model):
    name = models.CharField(max_length=64)
    git_url = models.URLField(
        null=True,
        help_text="The URL of the repo, not the clone URI",
    )

    def __str__(self):  # pragma: no cover
        return self.name


class ProjectVersion(models.Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    revision = models.IntegerField(default=0)
    commit_ref = models.CharField(
        max_length=64, help_text="Commit or tag representing the version"
    )
    bom_path = models.CharField(max_length=100)
    pcb_url = models.URLField(help_text="OSH park link to pcb", null=True, blank=True)
    parts = models.ManyToManyField(Part, related_name="projects", through="ProjectPart")
    pcb_cost = models.DecimalField(
        decimal_places=2, max_digits=6, help_text="for lot of 3", null=True, blank=True
    )
    synced = models.DateTimeField(null=True, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.project} v{self.revision}"

    @property
    def pcb_unit_cost(self):
        if self.pcb_cost is None:
            return 0
        return self.pcb_cost / 3

    @property
    def total_cost(self):
        return self.pcb_unit_cost + sum([p.line_cost for p in self.project_parts.all()])


class ProjectPart(models.Model):
    part = models.ForeignKey(
        Part,
        on_delete=models.PROTECT,
        related_name="project_parts",
        null=True,
        blank=True,
    )
    missing_part_description = models.CharField(max_length=256, null=True, blank=True)
    project_version = models.ForeignKey(
        ProjectVersion, on_delete=models.CASCADE, related_name="project_parts"
    )
    line_number = models.SmallIntegerField()
    quantity = models.SmallIntegerField()
    is_implicit = models.BooleanField(default=False)

    @property
    def line_cost(self):
        return self.part.unit_cost * self.quantity

    # derived field for footprint refs?


class ProjectPartFootprintRef(models.Model):
    project_part = models.ForeignKey(
        ProjectPart, on_delete=models.CASCADE, related_name="footprint_refs"
    )
    footprint_ref = models.CharField(max_length=8)

    def __str__(self):  # pragma: no cover
        return self.footprint_ref


class ProjectBuild(models.Model):
    project_version = models.ForeignKey(ProjectVersion, on_delete=models.PROTECT)
    quantity = models.SmallIntegerField()
    created = models.DateTimeField(default=timezone.now)
    cleared = models.DateTimeField(null=True, blank=True)
    completed = models.DateTimeField(null=True, blank=True)

    def __str__(self):  # pragma: no cover
        return f"{self.quantity}x {self.project_version}"


class ProjectBuildPartShortage(models.Model):
    part = models.ForeignKey(Part, on_delete=models.CASCADE)
    quantity = models.SmallIntegerField()
    project_build = models.ForeignKey(
        ProjectBuild, on_delete=models.CASCADE, related_name="shortfalls"
    )
    created = models.DateTimeField(default=timezone.now)


class ProjectBuildPartReservation(models.Model):
    """When clear to build, the parts for the build will be deducted from
    their respective inventory lines and reserved here"""

    inventory_action = models.ForeignKey(
        InventoryAction, on_delete=models.PROTECT, null=True
    )
    project_build = models.ForeignKey(
        ProjectBuild, on_delete=models.CASCADE, related_name="part_reservations"
    )
    created = models.DateTimeField(default=timezone.now)
    utilized = models.DateTimeField(null=True, blank=True)



from decimal import Decimal

import pytest


class TestPartModel:
    def test_unit_cost_no_vendor(self, part):
        assert part.unit_cost == 0

    def test_unit_cost(self, vendor_part):
        assert vendor_part.part.unit_cost == Decimal("0.0100")


class TestProjectVersionModel:
    def test_pcb_unit_cost_no_cost(self, project_version):
        project_version.pcb_cost = None
        project_version.save()
        assert project_version.pcb_unit_cost == 0

    def test_pcb_unit_cost(self, project_version):
        assert project_version.pcb_unit_cost == pytest.approx(14.23)

    def test_total_cost(self, project_version, project_part_factory, vendor_part):
        project_part_factory(vendor_part.part, project_version, quantity=102)
        print(project_version.project_parts.all()[0].quantity)
        print(project_version.project_parts.all()[0].part.unit_cost)
        project_version.refresh_from_db()
        assert float(project_version.total_cost) == pytest.approx(15.25)
